from pettingzoo.utils.deprecated_module import DeprecatedModule

basketball_pong_v0 = DeprecatedModule("basketball_pong", "v0", "v2")
basketball_pong_v1 = DeprecatedModule("basketball_pong", "v1", "v2")
boxing_v0 = DeprecatedModule("boxing", "v0", "v1")
combat_plane_v0 = DeprecatedModule("combat_plane", "v0", "v1")
combat_tank_v0 = DeprecatedModule("combat_tank", "v0", "v1")
double_dunk_v0 = DeprecatedModule("double_dunk", "v0", "v2")
double_dunk_v1 = DeprecatedModule("double_dunk", "v1", "v2")
entombed_competitive_v0 = DeprecatedModule("entombed_competitive", "v0", "v2")
entombed_competitive_v1 = DeprecatedModule("entombed_competitive", "v1", "v2")
entombed_cooperative_v0 = DeprecatedModule("entombed_cooperative", "v0", "v2")
entombed_cooperative_v1 = DeprecatedModule("entombed_cooperative", "v1", "v2")
flag_capture_v0 = DeprecatedModule("flag_capture", "v0", "v1")
foozpong_v0 = DeprecatedModule("foozpong", "v0", "v2")
foozpong_v1 = DeprecatedModule("foozpong", "v1", "v2")
ice_hockey_v0 = DeprecatedModule("ice_hockey", "v0", "v1")
joust_v0 = DeprecatedModule("joust", "v0", "v2")
joust_v1 = DeprecatedModule("joust", "v1", "v2")
mario_bros_v0 = DeprecatedModule("mario_bros", "v0", "v2")
mario_bros_v1 = DeprecatedModule("mario_bros", "v1", "v2")
maze_craze_v0 = DeprecatedModule("maze_craze", "v0", "v2")
maze_craze_v1 = DeprecatedModule("maze_craze", "v1", "v2")
othello_v0 = DeprecatedModule("othello", "v0", "v2")
othello_v1 = DeprecatedModule("othello", "v1", "v2")
pong_v0 = DeprecatedModule("pong", "v0", "v2")
pong_v1 = DeprecatedModule("pong", "v1", "v2")
quadrapong_v0 = DeprecatedModule("quadrapong", "v0", "v3")
quadrapong_v1 = DeprecatedModule("quadrapong", "v1", "v3")
quadrapong_v2 = DeprecatedModule("quadrapong", "v2", "v3")
space_invaders_v0 = DeprecatedModule("space_invaders", "v0", "v1")
space_war_v0 = DeprecatedModule("space_war", "v0", "v1")
surround_v0 = DeprecatedModule("surround", "v0", "v1")
tennis_v0 = DeprecatedModule("tennis", "v0", "v2")
tennis_v1 = DeprecatedModule("tennis", "v1", "v2")
video_checkers_v0 = DeprecatedModule("video_checkers", "v0", "v3")
video_checkers_v1 = DeprecatedModule("video_checkers", "v1", "v3")
video_checkers_v2 = DeprecatedModule("video_checkers", "v2", "v3")
volleyball_pong_v0 = DeprecatedModule("volleyball_pong", "v0", "v2")
volleyball_pong_v1 = DeprecatedModule("volleyball_pong", "v1", "v2")
warlords_v0 = DeprecatedModule("warlords", "v0", "v2")
warlords_v1 = DeprecatedModule("warlords", "v1", "v2")
wizard_of_wor_v0 = DeprecatedModule("wizard_of_wor", "v0", "v2")
wizard_of_wor_v1 = DeprecatedModule("wizard_of_wor", "v1", "v2")
