from pettingzoo.utils.deprecated_module import DeprecatedModule

prison_v0 = DeprecatedModule("prison", "v0", "v3")
prison_v1 = DeprecatedModule("prison", "v1", "v3")
prison_v2 = DeprecatedModule("prison", "v2", "v3")
prospector_v0 = DeprecatedModule("prospector", "v0", "v4")
prospector_v1 = DeprecatedModule("prospector", "v1", "v4")
prospector_v2 = DeprecatedModule("prospector", "v2", "v4")
prospector_v3 = DeprecatedModule("prospector", "v3", "v4")
pistonball_v0 = DeprecatedModule("pistonball", "v0", "v4")
pistonball_v1 = DeprecatedModule("pistonball", "v1", "v4")
pistonball_v2 = DeprecatedModule("pistonball", "v2", "v4")
pistonball_v3 = DeprecatedModule("pistonball", "v3", "v4")
cooperative_pong_v0 = DeprecatedModule("cooperative_pong", "v0", "v3")
cooperative_pong_v1 = DeprecatedModule("cooperative_pong", "v1", "v3")
knights_archers_zombies_v0 = DeprecatedModule("knights_archers_zombies", "v0", "v7")
knights_archers_zombies_v1 = DeprecatedModule("knights_archers_zombies", "v1", "v7")
knights_archers_zombies_v2 = DeprecatedModule("knights_archers_zombies", "v2", "v7")
knights_archers_zombies_v3 = DeprecatedModule("knights_archers_zombies", "v3", "v7")
knights_archers_zombies_v4 = DeprecatedModule("knights_archers_zombies", "v4", "v7")
knights_archers_zombies_v5 = DeprecatedModule("knights_archers_zombies", "v5", "v7")
knights_archers_zombies_v6 = DeprecatedModule("knights_archers_zombies", "v6", "v7")
