from pettingzoo.utils.deprecated_module import DeprecatedModule

backgammon_v0 = DeprecatedModule("backgammon", "v0", "v3")
backgammon_v1 = DeprecatedModule("backgammon", "v1", "v3")
backgammon_v2 = DeprecatedModule("backgammon", "v2", "v3")
checkers_v0 = DeprecatedModule("checkers", "v0", "v3")
checkers_v1 = DeprecatedModule("checkers", "v1", "v3")
checkers_v2 = DeprecatedModule("checkers", "v2", "v3")
chess_v0 = DeprecatedModule("chess", "v0", "v3")
chess_v1 = DeprecatedModule("chess", "v1", "v3")
chess_v2 = DeprecatedModule("chess", "v2", "v3")
connect_four_v0 = DeprecatedModule("connect_four", "v0", "v3")
connect_four_v1 = DeprecatedModule("connect_four", "v1", "v3")
connect_four_v2 = DeprecatedModule("connect_four", "v2", "v3")
dou_dizhu_v0 = DeprecatedModule("dou_dizhu", "v0", "v3")
dou_dizhu_v1 = DeprecatedModule("dou_dizhu", "v1", "v3")
dou_dizhu_v2 = DeprecatedModule("dou_dizhu", "v2", "v3")
gin_rummy_v0 = DeprecatedModule("gin_rummy", "v0", "v3")
gin_rummy_v1 = DeprecatedModule("gin_rummy", "v1", "v3")
gin_rummy_v2 = DeprecatedModule("gin_rummy", "v2", "v3")
go_v0 = DeprecatedModule("go", "v0", "v3")
go_v1 = DeprecatedModule("go", "v1", "v3")
go_v2 = DeprecatedModule("go", "v2", "v3")
hanabi_v0 = DeprecatedModule("hanabi", "v0", "v4")
hanabi_v1 = DeprecatedModule("hanabi", "v1", "v4")
hanabi_v2 = DeprecatedModule("hanabi", "v2", "v4")
hanabi_v3 = DeprecatedModule("hanabi", "v3", "v4")
leduc_holdem_v0 = DeprecatedModule("leduc_holdem", "v0", "v3")
leduc_holdem_v1 = DeprecatedModule("leduc_holdem", "v1", "v3")
leduc_holdem_v2 = DeprecatedModule("leduc_holdem", "v2", "v3")
mahjong_v0 = DeprecatedModule("mahjong", "v0", "v3")
mahjong_v1 = DeprecatedModule("mahjong", "v1", "v3")
mahjong_v2 = DeprecatedModule("mahjong", "v2", "v3")
rps_v0 = DeprecatedModule("rps", "v0", "v1")
rpsls_v0 = DeprecatedModule("rpsls", "v0", "v1")
texas_holdem_no_limit_v0 = DeprecatedModule("texas_holdem_no_limit", "v0", "v3")
texas_holdem_no_limit_v1 = DeprecatedModule("texas_holdem_no_limit", "v1", "v3")
texas_holdem_no_limit_v2 = DeprecatedModule("texas_holdem_no_limit", "v2", "v3")
texas_holdem_v0 = DeprecatedModule("texas_holdem", "v0", "v3")
texas_holdem_v1 = DeprecatedModule("texas_holdem", "v1", "v3")
texas_holdem_v2 = DeprecatedModule("texas_holdem", "v2", "v3")
tictactoe_v0 = DeprecatedModule("tictactoe", "v0", "v3")
tictactoe_v1 = DeprecatedModule("tictactoe", "v1", "v3")
tictactoe_v2 = DeprecatedModule("tictactoe", "v2", "v3")
uno_v0 = DeprecatedModule("uno_v0", "v0", "v3")
uno_v1 = DeprecatedModule("uno_v0", "v1", "v3")
uno_v2 = DeprecatedModule("uno_v0", "v2", "v3")
